from typing import Any, Union
from decimal import Decimal
from math import factorial


def PoissonPD_validate(k: Any, lambda_: Any) -> None:
    if not isinstance(k, int):
        raise TypeError("k value lambda_st be an integer")
    if not isinstance(lambda_, (int, float)):
        raise TypeError("λ data type unsupported")
    if lambda_ <= 0:
        raise ValueError("λ value out of domain")

def PoissonPD_calculate(k: int, lambda_: Decimal) -> Decimal:
    return (lambda_ ** k) / ( factorial(k) * lambda_.exp() )

def PoissonPD(k: int, lambda_: Union[int, float]) -> float:
    PoissonPD_validate(k, lambda_)
    return float(PoissonPD_calculate(k, Decimal(str(lambda_))))


def PoissonCD_calculate(k: int, lambda_: Decimal) -> Decimal:
    return sum(PoissonPD_calculate(i, lambda_) for i in range(k + 1))

def PoissonCD(k: int, lambda_: Union[int, float]) -> float:
    PoissonPD_validate(k, lambda_)
    return float(PoissonCD_calculate(k, Decimal(str(lambda_))))


def InvPoissonCD_validate(area: Any, lambda_: Any) -> None:
    if not isinstance(area, (int, float)):
        raise TypeError("area data type unsupported")
    if area < 0:
        raise ValueError("area value out of domain")
    if not isinstance(lambda_, (int, float)):
        raise TypeError("λ data type unsupported")
    if lambda_ <= 0:
        raise ValueError("λ value out of domain")

def InvPoissonCD_calculate(area: Decimal, lambda_: Decimal) -> int:
    tempK: int = 0
    while True:
        if PoissonCD_calculate(tempK, lambda_) >= area:
            break
        tempK += 1
    return tempK

def InvPoissonCD(area: Union[int, float], lambda_: Union[int, float]) -> int:
    InvPoissonCD_validate(area, lambda_)
    return InvPoissonCD_calculate(Decimal(str(area)), Decimal(str(lambda_)))




from typing import Union, Any
from math import erf, exp, sqrt, tau
from statistics import _normal_dist_inv_cdf #type:ignore

def NormalPD_validate(x: Any, mu: Any, sigma: Any) -> None:
    if not isinstance(x, (int, float)):
        raise TypeError("Input x value is invalid")
    if not isinstance(mu, (int, float)):
        raise TypeError("Input µ value is invalid")
    if (not sigma) or not isinstance(sigma, (int, float)):
        raise TypeError("Input σ value is invalid")

def NormalPD_calculate(x: float, mu: float, sigma: float) -> float:
    variance = sigma * sigma
    return exp((x - mu) ** 2 / (-2.0 * variance)) / sqrt(tau * variance)

def NormalPD(x: Union[int, float], mu: Union[int, float], sigma: Union[int, float]) -> float:
    NormalPD_validate(x, mu, sigma)
    return NormalPD_calculate(float(x), float(mu), float(sigma))


def NormalCD_calculate(x: float, mu: float, sigma: float) -> float:
    return 0.5 * (1.0 + erf((x - mu) / (sigma * sqrt(2.0))))

def NormalCD(x: Union[int, float], mu: Union[int, float], sigma: Union[int, float]) -> float:
    NormalPD_validate(x, sigma, mu)
    return NormalCD_calculate(float(x), float(mu), float(sigma))


def InvNormalCD_validate(p: Any, mu: Any, sigma: Any) -> None:
    if not isinstance(p, float) or not 0 <= p <= 1:
        raise TypeError("Input p value must be a positive decimal precentage")
    if not isinstance(mu, (int, float)):
        raise TypeError("Input µ value is invalid")
    if not sigma or not isinstance(sigma, (int, float)):
        raise TypeError("Input σ value is invalid")

def InvNormalCD(p: Union[int, float], mu: Union[int, float], sigma: Union[int, float]) -> float:
    InvNormalCD_validate(p, mu, sigma)
    return _normal_dist_inv_cdf(
        float(p),
        float(mu),
        float(sigma)
    ) #type:ignore




from typing import Any
from decimal import Decimal
from math import comb

def BinomialPD_validate(x: Any, n: Any, p: Any) -> None:
    if not isinstance(x, int):
        raise TypeError("x must be a non-negative integer")
    if x < 0:
        raise ValueError("x value out of domain")
    if not isinstance(n, int):
        raise TypeError("n must be a non-negative integer")
    if n < 0:
        raise ValueError("n value out of domain")
    if not isinstance(p, float):
        raise TypeError("p must be non-negative float")
    if not 0 <= p <= 1:
        raise ValueError("p value out of domain")

def BinomialPD_calculate(x: int, n: int, p: Decimal) -> Decimal:
    return comb(n, x) * (p ** x) * ((1 - p) ** (n - x))

def BinomialPD(x: int, n: int, p: float) -> Decimal:
    BinomialPD_validate(x, n, p)
    return BinomialPD_calculate(x, n, Decimal(str(p)))


def BinomialCD_calculate(k: int, n: int, p: Decimal) -> Decimal:
    # hello: str = 'hello'.
    return sum((BinomialPD_calculate(i, n, p) for i in range(k + 1)))

def BinomialCD(x: int, n: int, p: float) -> Decimal:
    BinomialPD_validate(x, n, p)
    return BinomialCD_calculate(x, n, Decimal(str(p)))


def InvBinomialCD_validate(y: Any, n: Any, p: Any) -> None:
    if not isinstance(y, float):
        raise TypeError("y must be non-negative float")
    if not 0 <= y <= 1:
        raise ValueError("y value out of domain")
    if not isinstance(n, int):
        raise TypeError("n must be a non-negative integer")
    if n < 0:
        raise ValueError("n value out of domain")
    if not isinstance(p, float):
        raise TypeError("p must be non-negative float")
    if not 0 <= p <= 1:
        raise ValueError("p value out of domain")

def InvBinomialCD_calculate(x: Decimal, n: int, p: Decimal) -> int:
    for i in range(n + 1):
        tempN = i - 1
        tempX = BinomialCD_calculate(tempN, n, p)
        if tempX == x:
            return i
        if tempX > x:
            return tempN

def InvBinomialCD(y: float, n: int, p: float) -> int:
    InvBinomialCD_validate(y, n, p)
    return InvBinomialCD_calculate(Decimal(str(y)), n, Decimal(str(p)))




from typing import Any, Union
from decimal import Decimal

def GeometricPD(x: int, p: Union[int, float]) -> Decimal:
    GeometricPD_validate(x, p)
    return GeometricPD_calculate(x, Decimal(str(p)))

def GeometricPD_validate(x: Any, p: Any) -> None:
    if not isinstance(x, int):
        raise TypeError("x value must be a non-negative integer")
    if x < 0:
        raise ValueError("x value out of domain")
    if not isinstance(p, float):
        raise TypeError("p value must be non-negative float")
    if not 0 <= p <= 1:
        raise ValueError("p value out of domain")

def GeometricPD_calculate(x: int, p: Decimal) -> Decimal:
    return ((1 - p) ** (x - 1)) * p


def GeometricCD(x: int, p: float) -> Decimal:
    GeometricPD_validate(x, p)
    return GeometricCD_calculate(x, Decimal(str(p)))

def GeometricCD_calculate(x: int, p: Decimal) -> Decimal:
    return 1 - (1 - p) ** x


def InvGeometricCD(area: float, p: float) -> int:
    InvGeometricCD_validate(area, p)
    return InvGeometricCD_calculate(
        Decimal(str(area)),
        Decimal(str(p))
    )

def InvGeometricCD_validate(area: Any, p: Any) -> None:
    if not isinstance(area, float):
        raise TypeError("area value must be non-negative float")
    if not 0 <= area <= 1:
        raise ValueError("area value out of domain")
    if not isinstance(p, float):
        raise TypeError("p value must be non-negative float")
    if not 0 <= p <= 1:
        raise ValueError("p value out of domain")

def InvGeometricCD_calculate(area: Decimal, p: Decimal) -> int:
    cumulative = Decimal()
    i = 1
    while cumulative <= area:
        cumulative += GeometricPD_calculate(i, p)
        i += 1
    return i - 1




from .distributions_binomial import BinomialPD, BinomialCD, InvBinomialCD
from .distributions_poisson import PoissonPD, PoissonCD, InvPoissonCD
from .distributions_normal import NormalPD, NormalCD, InvNormalCD
from .distributions_geometric import GeometricPD, GeometricCD, InvGeometricCD

__all__ = (
    'BinomialPD',
    'BinomialCD',
    'InvBinomialCD',
    'PoissonPD',
    'PoissonCD',
    'InvPoissonCD',
    'NormalPD',
    'NormalCD',
    'InvNormalCD',
    'GeometricPD',
    'GeometricCD',
    'InvGeometricCD'
)
